from collections import namedtuple
from enum import Enum
from typing import Any, List, Optional


class PrimitiveCategory(Enum):
    VOID = "VOID"
    BOOLEAN = "BOOLEAN"
    BYTE = "BYTE"
    SHORT = "SHORT"
    INT = "INT"
    LONG = "LONG"
    FLOAT = "FLOAT"
    DOUBLE = "DOUBLE"
    STRING = "STRING"
    DATE = "DATE"
    TIMESTAMP = "TIMESTAMP"
    TIMESTAMPLOCALTZ = "TIMESTAMPLOCALTZ"
    BINARY = "BINARY"
    DECIMAL = "DECIMAL"
    VARCHAR = "VARCHAR"
    CHAR = "CHAR"
    INTERVAL_YEAR_MONTH = "INTERVAL_YEAR_MONTH"
    INTERVAL_DAY_TIME = "INTERVAL_DAY_TIME"
    UNKNOWN = "UNKNOWN"


class HTypeCategory(Enum):
    PRIMITIVE = PrimitiveCategory
    STRUCT = "STRUCT"
    MAP = "MAP"
    LIST = "LIST"
    UNION = "UNION"


class HType:
    def __init__(self, name: str, category: HTypeCategory):
        self.name = name
        self.category = category


class HPrimitiveType(HType):
    def __init__(self, primitive_type: PrimitiveCategory):
        super().__init__(primitive_type.value, HTypeCategory.PRIMITIVE)
        self.primitive_type = primitive_type


class HMapType(HType):
    def __init__(self, key_type: HType, value_type: HType):
        super().__init__("MAP", HTypeCategory.MAP)
        self.key_type = key_type
        self.value_type = value_type


class HListType(HType):
    def __init__(self, element_type: HType):
        super().__init__("LIST", HTypeCategory.LIST)
        self.element_type = element_type


class HUnionType(HType):
    def __init__(self, types: List[HType]):
        super().__init__("UNION", HTypeCategory.UNION)
        self.types = types


class HVarcharType(HType):
    def __init__(self, length: int):
        MAX_VARCHAR_LENGTH = 65535
        if length > MAX_VARCHAR_LENGTH:
            raise ValueError("Varchar length cannot exceed 65535")
        super().__init__("VARCHAR", HTypeCategory.PRIMITIVE)
        self.length = length


class HCharType(HType):
    def __init__(self, length: int):
        MAX_CHAR_LENGTH = 255
        if length > MAX_CHAR_LENGTH:
            raise ValueError("Char length cannot exceed 255")
        super().__init__("CHAR", HTypeCategory.PRIMITIVE)
        self.length = length


class HDecimalType(HType):
    def __init__(self, precision: int, scale: int):
        MAX_PRECISION = 38
        MAX_SCALE = 38
        if precision > MAX_PRECISION:
            raise ValueError("Decimal precision cannot exceed 38")
        if scale > MAX_SCALE:
            raise ValueError("Decimal scale cannot exceed 38")
        super().__init__("DECIMAL", HTypeCategory.PRIMITIVE)
        self.precision = precision
        self.scale = scale


class HStructType(HType):
    def __init__(self, names: List[str], types: List[HType]):
        if len(names) != len(types):
            raise ValueError("mismatched size of names and types.")
        if None in names:
            raise ValueError("names cannot contain None")
        if None in types:
            raise ValueError("types cannot contain None")

        super().__init__("STRUCT", HTypeCategory.STRUCT)
        self.names = names
        self.types = types


# We need to maintain a list of the expected serialized type names.
# Thrift will return the type in string format and we will have to parse it to get the type.
class SerdeTypeNameConstants(Enum):
    # Primitive types
    VOID = "void"
    BOOLEAN = "boolean"
    TINYINT = "tinyint"
    SMALLINT = "smallint"
    INT = "int"
    BIGINT = "bigint"
    FLOAT = "float"
    DOUBLE = "double"
    STRING = "string"
    DATE = "date"
    CHAR = "char"
    VARCHAR = "varchar"
    TIMESTAMP = "timestamp"
    TIMESTAMPLOCALTZ = "timestamp with local time zone"
    DECIMAL = "decimal"
    BINARY = "binary"
    INTERVAL_YEAR_MONTH = "interval_year_month"
    INTERVAL_DAY_TIME = "interval_day_time"
    # Complex types
    LIST = "array"
    MAP = "map"
    STRUCT = "struct"
    UNION = "uniontype"
    UNKNOWN = "unknown"


# We need a mapping between the two definitions of primitive types.
# The first is the one used by Hive and the second is the one used by Thrift returned values.
primitive_to_serde_mapping = {
    PrimitiveCategory.VOID: SerdeTypeNameConstants.VOID,
    PrimitiveCategory.BOOLEAN: SerdeTypeNameConstants.BOOLEAN,
    PrimitiveCategory.BYTE: SerdeTypeNameConstants.TINYINT,
    PrimitiveCategory.SHORT: SerdeTypeNameConstants.SMALLINT,
    PrimitiveCategory.INT: SerdeTypeNameConstants.INT,
    PrimitiveCategory.LONG: SerdeTypeNameConstants.BIGINT,
    PrimitiveCategory.FLOAT: SerdeTypeNameConstants.FLOAT,
    PrimitiveCategory.DOUBLE: SerdeTypeNameConstants.DOUBLE,
    PrimitiveCategory.STRING: SerdeTypeNameConstants.STRING,
    PrimitiveCategory.DATE: SerdeTypeNameConstants.DATE,
    PrimitiveCategory.TIMESTAMP: SerdeTypeNameConstants.TIMESTAMP,
    PrimitiveCategory.TIMESTAMPLOCALTZ: SerdeTypeNameConstants.TIMESTAMPLOCALTZ,
    PrimitiveCategory.BINARY: SerdeTypeNameConstants.BINARY,
    PrimitiveCategory.DECIMAL: SerdeTypeNameConstants.DECIMAL,
    PrimitiveCategory.VARCHAR: SerdeTypeNameConstants.VARCHAR,
    PrimitiveCategory.CHAR: SerdeTypeNameConstants.CHAR,
    PrimitiveCategory.INTERVAL_YEAR_MONTH: SerdeTypeNameConstants.INTERVAL_YEAR_MONTH,
    PrimitiveCategory.INTERVAL_DAY_TIME: SerdeTypeNameConstants.INTERVAL_DAY_TIME,
    PrimitiveCategory.UNKNOWN: SerdeTypeNameConstants.UNKNOWN,
}

# We also need the inverse though.
serde_to_primitive_mapping = {v.name: k for k, v in primitive_to_serde_mapping.items()}


# This function gets us the serde type name from the primitive one.
def primitive_to_serde(primitive_category: PrimitiveCategory) -> SerdeTypeNameConstants:
    return primitive_to_serde_mapping[primitive_category]


# This function gets us the primitive type name from the serde one.
def serde_to_primitive(serde_type_name: str) -> Optional[PrimitiveCategory]:
    return serde_to_primitive_mapping.get(serde_type_name)


# We also need a way to get the serde enum key from the string we get from Thrift.
def get_serde_type_by_value(value):
    for member in SerdeTypeNameConstants:
        if member.value == value:
            return member.name
    return None


# To be used for tokening the Thrift type string. We need to tokenize the
# string to get the type name and the type parameters.
class Token(namedtuple("Token", ["position", "text", "type"])):
    def __new__(cls, position: int, text: str, type_: bool) -> Any:
        if text is None:
            raise ValueError("text is null")
        return super().__new__(cls, position, text, type_)

    def __str__(self):
        return f"{self.position}:{self.text}"


class PrimitiveParts(namedtuple("PrimitiveParts", ["type_name", "type_params"])):
    def __new__(cls, type_name: str, type_params: List[str]) -> Any:
        if type_name is None:
            raise ValueError("type_name is null")
        if type_params is None:
            raise ValueError("type_params is null")
        return super().__new__(cls, type_name, type_params)

    def __str__(self):
        return f"{self.type_name}:{self.type_params}"


# Util functions to parse the Thrift column types (as strings) and return the
# expected hive type
class TypeParser:
    def __init__(self, type_info_string: str):
        self.type_string = type_info_string
        self.type_tokens = self.tokenize(type_info_string)
        self.index = 0

    # we want the base name of the type, in general the format of a
    # parameterized type is <base_name>(<type1>, <type2>, ...), so the base
    # name is the string before the first '('
    @staticmethod
    def get_base_name(type_name: str) -> str:
        index = type_name.find("(")
        if index == -1:
            return type_name
        return type_name[:index]

    # Is the type named using the allowed characters?
    @staticmethod
    def is_valid_type_char(c: str) -> bool:
        return c.isalnum() or c in ("_", ".", " ", "$")

    # The concept of the tokenizer is to split the type string into tokens. The
    # tokens are either type names or type parameters. The type parameters are
    # enclosed in parentheses.
    def tokenize(self, type_info_string: str) -> List[Token]:
        tokens = []
        begin = 0
        end = 1
        while end <= len(type_info_string):
            if (
                begin > 0
                and type_info_string[begin - 1] == "("
                and type_info_string[begin] == "'"
            ):
                begin += 1
                end += 1
                while type_info_string[end] != "'":
                    end += 1

            elif type_info_string[begin] == "'" and type_info_string[begin + 1] == ")":
                begin += 1
                end += 1

            if (
                end == len(type_info_string)
                or not self.is_valid_type_char(type_info_string[end - 1])
                or not self.is_valid_type_char(type_info_string[end])
            ):
                token = Token(
                    begin,
                    type_info_string[begin:end],
                    self.is_valid_type_char(type_info_string[begin]),
                )
                tokens.append(token)
                begin = end
            end += 1

        return tokens

    def peek(self) -> Optional[Token]:
        if self.index >= len(self.type_tokens):
            return None
        return self.type_tokens[self.index]

    def expect(self, item: str, alternative: Optional[str] = None) -> Token:
        if self.index >= len(self.type_tokens):
            raise ValueError(f"Error: {item} expected at the end of 'typeInfoString'")

        token = self.type_tokens[self.index]

        if item == "type":
            if (
                token.text
                not in [
                    HTypeCategory.LIST.value,
                    HTypeCategory.MAP.value,
                    HTypeCategory.STRUCT.value,
                    HTypeCategory.UNION.value,
                ]
                and (self.get_base_name(token.text) is None)
                and (token.text != alternative)
            ):
                raise ValueError(
                    f"Error: {item} expected at the position {token.position} "
                    f"of 'typeInfoString' but '{token.text}' is found."
                )
        elif item == "name":
            if not token.type and token.text != alternative:
                raise ValueError(
                    f"Error: {item} expected at the position {token.position} "
                    f"of 'typeInfoString' but '{token.text}' is found."
                )
        elif item != token.text and token.text != alternative:
            raise ValueError(
                f"Error: {item} expected at the position {token.position} of "
                f"'typeInfoString' but '{token.text}' is found."
            )

        self.index += 1  # increment the index
        return token

    def parse_params(self) -> List[str]:
        params = []

        token: Optional[Token] = self.peek()
        if token is not None and token.text == "(":
            token = self.expect("(", None)
            token = self.peek()
            while token is None or token.text != ")":
                token = self.expect("name", None)
                params.append(token.text)
                token = self.expect(",", ")")

            if not params:
                raise ValueError(
                    "type parameters expected for type string 'typeInfoString'"
                )

        return params

    def parse_type(self) -> HType:
        token = self.expect("type")

        # first we take care of primitive types.
        serde_val = get_serde_type_by_value(token.text)
        if serde_val and serde_val not in (
            HTypeCategory.LIST.value,
            HTypeCategory.MAP.value,
            HTypeCategory.STRUCT.value,
            HTypeCategory.UNION.value,
        ):
            primitive_type = serde_to_primitive(serde_val)
            if primitive_type and primitive_type is not PrimitiveCategory.UNKNOWN:
                params = self.parse_params()
                if primitive_type in (
                    PrimitiveCategory.CHAR,
                    PrimitiveCategory.VARCHAR,
                ):
                    if len(params) == 0:
                        raise ValueError(
                            "char/varchar type must have a length specified"
                        )
                    if len(params) == 1:
                        length = int(params[0])
                        if primitive_type is PrimitiveCategory.CHAR:
                            return HCharType(length)
                        elif primitive_type is PrimitiveCategory.VARCHAR:
                            return HVarcharType(length)
                    else:
                        raise ValueError(
                            f"Error: {token.text} type takes only one parameter, but instead "
                            f"{len(params)} parameters are found."
                        )
                elif primitive_type is PrimitiveCategory.DECIMAL:
                    if len(params) == 0:
                        return HDecimalType(10, 0)
                    elif len(params) == 1:
                        precision = int(params[0])
                        return HDecimalType(precision, 0)
                    elif len(params) == 2:
                        precision = int(params[0])
                        scale = int(params[1])
                        return HDecimalType(precision, scale)
                    else:
                        raise ValueError(
                            f"Error: {token.text} type takes only two parameters, but instead "
                            f"{len(params)} parameters are found."
                        )
                else:
                    return HPrimitiveType(primitive_type)

        # next we take care of complex types.
        if HTypeCategory.LIST.value == serde_val:
            self.expect("<")
            element_type = self.parse_type()
            self.expect(">")
            return HListType(element_type)

        if HTypeCategory.MAP.value == serde_val:
            self.expect("<")
            key_type = self.parse_type()
            self.expect(",")
            value_type = self.parse_type()
            self.expect(">")
            return HMapType(key_type, value_type)

        if HTypeCategory.STRUCT.value == serde_val:
            self.expect("<")
            names = []
            types = []
            token: Optional[Token] = self.peek()
            while token is not None and token.text != ">":
                field_name = self.expect("name")
                self.expect(":")
                field_type = self.parse_type()
                names.append(field_name.text)
                types.append(field_type)
                token: Optional[Token] = self.peek()
                if token is not None and token.text == ",":
                    self.expect(",")
                    token = self.peek()
            self.expect(">")
            return HStructType(names, types)

        if HTypeCategory.UNION.value == serde_val:
            self.expect("<")
            types = []
            token = self.peek()
            while token is not None and token.text != ">":
                field_type = self.parse_type()
                types.append(field_type)
                token = self.peek()
                if token is not None and token.text == ",":
                    self.expect(",")
                    token = self.peek()
            self.expect(">")
            return HUnionType(types)

        # if we reach this point and we haven't figure out what to do, then we
        # better raise an error.
        raise ValueError(f"Error: {token.text} is not a valid type.")



import pytest

from pymetastore.htypes import (
    HCharType,
    HDecimalType,
    HListType,
    HMapType,
    HPrimitiveType,
    HStructType,
    HTypeCategory,
    HUnionType,
    HVarcharType,
    PrimitiveCategory,
    SerdeTypeNameConstants,
    Token,
    TypeParser,
)

# Although type systems change, it is a rare event and even if it happens, we
# should be very careful as way too many things can break. For this reason
# we'll be testing values and members for the types we define to catch any
# changes. For this reason we'll be primarily be testing the values and not the
# utility functions for the conversion between types. These functions are
# trivial and they primarily expose a thin API over the mappings we define.


# tests for primitive types.
def test_primitivecategory_members():
    expected_members = {
        "VOID",
        "BOOLEAN",
        "BYTE",
        "SHORT",
        "INT",
        "LONG",
        "FLOAT",
        "DOUBLE",
        "STRING",
        "DATE",
        "TIMESTAMP",
        "TIMESTAMPLOCALTZ",
        "BINARY",
        "DECIMAL",
        "VARCHAR",
        "CHAR",
        "INTERVAL_YEAR_MONTH",
        "INTERVAL_DAY_TIME",
        "UNKNOWN",
    }

    actual_members = {member.name for member in PrimitiveCategory}
    assert actual_members == expected_members


def test_primitivecategory_values():
    expected_values = {
        "VOID",
        "BOOLEAN",
        "BYTE",
        "SHORT",
        "INT",
        "LONG",
        "FLOAT",
        "DOUBLE",
        "STRING",
        "DATE",
        "TIMESTAMP",
        "TIMESTAMPLOCALTZ",
        "BINARY",
        "DECIMAL",
        "VARCHAR",
        "CHAR",
        "INTERVAL_YEAR_MONTH",
        "INTERVAL_DAY_TIME",
        "UNKNOWN",
    }

    actual_values = {member.value for member in PrimitiveCategory}
    assert actual_values == expected_values


# tests for HType definitions.
def test_htype_members():
    expected_members = {"PRIMITIVE", "STRUCT", "MAP", "LIST", "UNION"}

    actual_members = {member.name for member in HTypeCategory}
    assert actual_members == expected_members


def test_htype_values():
    expected_values = {PrimitiveCategory, "STRUCT", "MAP", "LIST", "UNION"}

    actual_values = {member.value for member in HTypeCategory}
    assert actual_values == expected_values


# Tests for the Serde definitions. Remember that we are converting between
# Thrift and Hive types so we need to account for that. It's important to test
# that stuff here because the thrift types are auto generated and we should
# catch any changes that might happen in the .thrift files as soon as possible.
def test_serdetypes_members():
    expected_members = {
        "VOID",
        "BOOLEAN",
        "TINYINT",
        "SMALLINT",
        "INT",
        "BIGINT",
        "FLOAT",
        "DOUBLE",
        "STRING",
        "DATE",
        "CHAR",
        "VARCHAR",
        "TIMESTAMP",
        "TIMESTAMPLOCALTZ",
        "DECIMAL",
        "BINARY",
        "INTERVAL_YEAR_MONTH",
        "INTERVAL_DAY_TIME",
        "LIST",
        "MAP",
        "STRUCT",
        "UNION",
        "UNKNOWN",
    }

    actual_members = {member.name for member in SerdeTypeNameConstants}
    assert actual_members == expected_members


def test_serdetypes_values():
    expected_values = {
        "void",
        "boolean",
        "tinyint",
        "smallint",
        "int",
        "bigint",
        "float",
        "double",
        "string",
        "date",
        "char",
        "varchar",
        "timestamp",
        "timestamp with local time zone",
        "decimal",
        "binary",
        "interval_year_month",
        "interval_day_time",
        "array",
        "map",
        "struct",
        "uniontype",
        "unknown",
    }

    actual_values = {member.value for member in SerdeTypeNameConstants}
    assert actual_values == expected_values


def test_token_creation():
    # Test successful creation of Token instance
    token = Token(position=0, text="double", type_=True)
    assert token.position == 0
    assert token.text == "double"
    assert token.type is True

    # Test the string representation of the Token instance
    assert str(token) == "0:double"

    # Test failure when providing None as text
    with pytest.raises(ValueError):
        Token(position=0, text=None, type_=True)


def test_token_properties():
    token = Token(position=1, text="<", type_=False)

    # Test the properties (position, text, type_) of the Token instance
    assert token.position == 1
    assert token.text == "<"
    assert token.type is False

    # Test the string representation of the Token instance
    assert str(token) == "1:<"


# Tests for the TypeParser class
def test_typeparser_tokenize():
    parser = TypeParser("STRING")
    assert parser.index == 0
    assert parser.type_tokens == [Token(position=0, text="STRING", type_=True)]
    parser = TypeParser("STRUCT<a:INT,b:STRING>")
    assert parser.index == 0

    assert parser.type_tokens == [
        Token(position=0, text="STRUCT", type_=True),
        Token(position=6, text="<", type_=False),
        Token(position=7, text="a", type_=True),
        Token(position=8, text=":", type_=False),
        Token(position=9, text="INT", type_=True),
        Token(position=12, text=",", type_=False),
        Token(position=13, text="b", type_=True),
        Token(position=14, text=":", type_=False),
        Token(position=15, text="STRING", type_=True),
        Token(position=21, text=">", type_=False),
    ]

    parser = TypeParser("MAP<STRING,STRUCT<a:INT,b:STRING>>")
    assert parser.index == 0
    assert parser.type_tokens == [
        Token(position=0, text="MAP", type_=True),
        Token(position=3, text="<", type_=False),
        Token(position=4, text="STRING", type_=True),
        Token(position=10, text=",", type_=False),
        Token(position=11, text="STRUCT", type_=True),
        Token(position=17, text="<", type_=False),
        Token(position=18, text="a", type_=True),
        Token(position=19, text=":", type_=False),
        Token(position=20, text="INT", type_=True),
        Token(position=23, text=",", type_=False),
        Token(position=24, text="b", type_=True),
        Token(position=25, text=":", type_=False),
        Token(position=26, text="STRING", type_=True),
        Token(position=32, text=">", type_=False),
        Token(position=33, text=">", type_=False),
    ]

    parser = TypeParser("UNIONTYPE<int,double,array<string>,struct<a:int,b:string>>")
    assert parser.index == 0

    assert parser.type_tokens == [
        Token(position=0, text="UNIONTYPE", type_=True),
        Token(position=9, text="<", type_=False),
        Token(position=10, text="int", type_=True),
        Token(position=13, text=",", type_=False),
        Token(position=14, text="double", type_=True),
        Token(position=20, text=",", type_=False),
        Token(position=21, text="array", type_=True),
        Token(position=26, text="<", type_=False),
        Token(position=27, text="string", type_=True),
        Token(position=33, text=">", type_=False),
        Token(position=34, text=",", type_=False),
        Token(position=35, text="struct", type_=True),
        Token(position=41, text="<", type_=False),
        Token(position=42, text="a", type_=True),
        Token(position=43, text=":", type_=False),
        Token(position=44, text="int", type_=True),
        Token(position=47, text=",", type_=False),
        Token(position=48, text="b", type_=True),
        Token(position=49, text=":", type_=False),
        Token(position=50, text="string", type_=True),
        Token(position=56, text=">", type_=False),
        Token(position=57, text=">", type_=False),
    ]
    parser = TypeParser("uniontype<integer,string>")
    assert parser.index == 0
    assert parser.type_tokens == [
        Token(position=0, text="uniontype", type_=True),
        Token(position=9, text="<", type_=False),
        Token(position=10, text="integer", type_=True),
        Token(position=17, text=",", type_=False),
        Token(position=18, text="string", type_=True),
        Token(position=24, text=">", type_=False),
    ]

    parser = TypeParser(
        "STRUCT<houseno:STRING,streetname:STRING,town:STRING,postcode:STRING>"
    )
    assert parser.index == 0
    assert parser.type_tokens == [
        Token(position=0, text="STRUCT", type_=True),
        Token(position=6, text="<", type_=False),
        Token(position=7, text="houseno", type_=True),
        Token(position=14, text=":", type_=False),
        Token(position=15, text="STRING", type_=True),
        Token(position=21, text=",", type_=False),
        Token(position=22, text="streetname", type_=True),
        Token(position=32, text=":", type_=False),
        Token(position=33, text="STRING", type_=True),
        Token(position=39, text=",", type_=False),
        Token(position=40, text="town", type_=True),
        Token(position=44, text=":", type_=False),
        Token(position=45, text="STRING", type_=True),
        Token(position=51, text=",", type_=False),
        Token(position=52, text="postcode", type_=True),
        Token(position=60, text=":", type_=False),
        Token(position=61, text="STRING", type_=True),
        Token(position=67, text=">", type_=False),
    ]


# Basic Types
def test_basic_type_parser():
    # String Type
    parser = TypeParser("string")
    ptype = parser.parse_type()
    assert ptype.name == HPrimitiveType(PrimitiveCategory.STRING).name
    assert ptype.category == HPrimitiveType(PrimitiveCategory.STRING).category

    # Int Type
    parser = TypeParser("int")
    ptype = parser.parse_type()
    assert ptype.name == HPrimitiveType(PrimitiveCategory.INT).name
    assert ptype.category == HPrimitiveType(PrimitiveCategory.INT).category

    # Long Type
    parser = TypeParser("double")
    ptype = parser.parse_type()
    assert ptype.name == HPrimitiveType(PrimitiveCategory.DOUBLE).name
    assert ptype.category == HPrimitiveType(PrimitiveCategory.DOUBLE).category

    # Float Type
    parser = TypeParser("float")
    ptype = parser.parse_type()
    assert ptype.name == HPrimitiveType(PrimitiveCategory.FLOAT).name
    assert ptype.category == HPrimitiveType(PrimitiveCategory.FLOAT).category

    # Boolean Type
    parser = TypeParser("boolean")
    ptype = parser.parse_type()
    assert ptype.name == HPrimitiveType(PrimitiveCategory.BOOLEAN).name
    assert ptype.category == HPrimitiveType(PrimitiveCategory.BOOLEAN).category

    # Binary Type
    parser = TypeParser("binary")
    ptype = parser.parse_type()
    assert ptype.name == HPrimitiveType(PrimitiveCategory.BINARY).name
    assert ptype.category == HPrimitiveType(PrimitiveCategory.BINARY).category

    # Timestamp Type
    parser = TypeParser("timestamp")
    ptype = parser.parse_type()
    assert ptype.name == HPrimitiveType(PrimitiveCategory.TIMESTAMP).name
    assert ptype.category == HPrimitiveType(PrimitiveCategory.TIMESTAMP).category

    # Date Type
    parser = TypeParser("date")
    ptype = parser.parse_type()
    assert ptype.name == HPrimitiveType(PrimitiveCategory.DATE).name
    assert ptype.category == HPrimitiveType(PrimitiveCategory.DATE).category


# Decimal Type
def test_decimal_type_parser():
    parser = TypeParser("decimal")
    ptype = parser.parse_type()

    # default precision and scale decimal
    assertd = HDecimalType(10, 0)
    assert isinstance(ptype, HDecimalType)
    assert ptype.precision == assertd.precision
    assert ptype.scale == assertd.scale

    # with precision defined only
    parser = TypeParser("decimal(20)")
    ptype = parser.parse_type()
    assertd = HDecimalType(20, 0)
    assert isinstance(ptype, HDecimalType)
    assert ptype.precision == assertd.precision
    assert ptype.scale == assertd.scale

    # with precision and scale defined
    parser = TypeParser("decimal(20,10)")
    ptype = parser.parse_type()
    assertd = HDecimalType(20, 10)
    assert isinstance(ptype, HDecimalType)
    assert ptype.precision == assertd.precision
    assert ptype.scale == assertd.scale

    # Test accepted boundaries for parameters
    parser = TypeParser("decimal(39,0)")
    with pytest.raises(ValueError, match="Decimal precision cannot exceed 38"):
        parser.parse_type()

    # Test accepted boundaries for parameters
    parser = TypeParser("decimal(37,40)")
    with pytest.raises(ValueError, match="Decimal scale cannot exceed 38"):
        parser.parse_type()

    # Test accepted boundaries for parameters
    parser = TypeParser("decimal(39,40)")
    with pytest.raises(ValueError, match="Decimal precision cannot exceed 38"):
        parser.parse_type()


# CHAR Type
def test_char_type_parser():
    parser = TypeParser("char(10)")
    ptype = parser.parse_type()
    assert isinstance(ptype, HCharType)
    assert ptype.length == 10

    # Testing char without length
    parser = TypeParser("char")
    with pytest.raises(
        ValueError, match="char/varchar type must have a length specified"
    ):
        parser.parse_type()

    # Testing char with more than 255 length
    parser = TypeParser("char(65536)")
    with pytest.raises(ValueError, match="Char length cannot exceed 255"):
        parser.parse_type()

    # Testing char with more than one parameter
    parser = TypeParser("char(10,20)")
    with pytest.raises(
        ValueError,
        match="Error: char type takes only one parameter, but instead 2 parameters are found.",
    ):
        parser.parse_type()

    # VARCHAR Type


def test_var_char_type_parser():
    parser = TypeParser("varchar(10)")
    ptype = parser.parse_type()
    assert isinstance(ptype, HVarcharType)
    assert ptype.length == 10

    # Testing varchar without length
    parser = TypeParser("varchar")
    with pytest.raises(
        ValueError, match="char/varchar type must have a length specified"
    ):
        parser.parse_type()

    # Testing varchar with more than 65535 length
    parser = TypeParser("varchar(65536)")
    with pytest.raises(ValueError, match="Varchar length cannot exceed 65535"):
        parser.parse_type()

    # Testing varchar with more than one parameter
    parser = TypeParser("varchar(10,20)")
    with pytest.raises(
        ValueError,
        match="Error: varchar type takes only one parameter, but instead 2 parameters are found.",
    ):
        parser.parse_type()


# ARRAY Type
def test_array_type_parser():
    parser = TypeParser("array<int>")
    ptype = parser.parse_type()
    assert isinstance(ptype, HListType)
    assert ptype.element_type.name == "INT"
    assert (
        ptype.element_type.category.PRIMITIVE
        == HPrimitiveType(PrimitiveCategory.INT).category
    )


def test_struct_type_parser():
    parser = TypeParser("struct<name:string,age:int>")
    ptype: HStructType = parser.parse_type()
    assert isinstance(ptype, HStructType)
    assert ptype.names == ["name", "age"]
    assert ptype.types[0].category == HPrimitiveType(PrimitiveCategory.STRING).category
    assert ptype.types[0].name == HPrimitiveType(PrimitiveCategory.STRING).name
    assert ptype.types[1].category == HPrimitiveType(PrimitiveCategory.INT).category
    assert ptype.types[1].name == HPrimitiveType(PrimitiveCategory.INT).name


def test_map_type_parser():
    parser = TypeParser("map<string,int>")
    ptype = parser.parse_type()
    assert isinstance(ptype, HMapType)
    assert ptype.key_type.name == "STRING"
    assert ptype.key_type.category == HPrimitiveType(PrimitiveCategory.STRING).category
    assert ptype.value_type.name == "INT"
    assert ptype.value_type.category == HPrimitiveType(PrimitiveCategory.INT).category


def test_union_type_parser():
    parser = TypeParser("uniontype<int,string>")
    ptype = parser.parse_type()
    assert isinstance(ptype, HUnionType)
    assert ptype.types[0].name == "INT"
    assert ptype.types[0].category == HPrimitiveType(PrimitiveCategory.INT).category
    assert ptype.types[1].name == "STRING"
    assert ptype.types[1].category == HPrimitiveType(PrimitiveCategory.STRING).category



import os

import pytest
from thrift.protocol import TBinaryProtocol
from thrift.transport import TSocket, TTransport

from pymetastore.hive_metastore import ttypes
from pymetastore.hive_metastore.ThriftHiveMetastore import Client
from pymetastore.htypes import (
    HCharType,
    HDecimalType,
    HListType,
    HMapType,
    HPrimitiveType,
    HStructType,
    HType,
    HTypeCategory,
    HUnionType,
    HVarcharType,
)
from pymetastore.metastore import (
    HMS,
    BucketingVersion,
    HColumn,
    HDatabase,
    HPartition,
    HStorage,
    HTable,
    HiveBucketProperty,
    StorageFormat,
)


@pytest.fixture(scope="module")
def hive_client():
    host = os.environ.get("HMS_HOST", "localhost")
    port = int(os.environ.get("HMS_PORT", 9083))
    transport = TSocket.TSocket(host, port)
    transport = TTransport.TBufferedTransport(transport)
    protocol = TBinaryProtocol.TBinaryProtocol(transport)
    client = Client(protocol)
    transport.open()

    yield client

    transport.close()


@pytest.fixture(scope="module", autouse=True)
def setup_data(hive_client):
    db = ttypes.Database(
        name="test_db",
        description="This is a test database",
        locationUri="/tmp/test_db.db",
        parameters={},
        ownerName="owner",
    )

    if "test_db" in hive_client.get_all_databases():
        hive_client.drop_database("test_db", True, True)
    hive_client.create_database(db)

    cols = [
        ttypes.FieldSchema(name="col1", type="int", comment="c1"),
        ttypes.FieldSchema(name="col2", type="string", comment="c2"),
    ]

    partitionKeys = [ttypes.FieldSchema(name="partition", type="string", comment="")]

    serde_info = ttypes.SerDeInfo(
        name="test_serde",
        serializationLib="org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe",
        parameters={"field.delim": ","},
    )

    storageDesc = ttypes.StorageDescriptor(
        location="/tmp/test_db/test_table",
        cols=cols,
        inputFormat="org.apache.hadoop.mapred.TextInputFormat",
        outputFormat="org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat",
        compressed=False,
        numBuckets=-1,
        serdeInfo=serde_info,
        bucketCols=[],
    )

    table = ttypes.Table(
        tableName="test_table",
        dbName="test_db",
        owner="owner",
        createTime=0,
        lastAccessTime=0,
        retention=0,
        sd=storageDesc,
        partitionKeys=partitionKeys,
        parameters={},
        tableType="EXTERNAL_TABLE",
    )

    if "test_table" in hive_client.get_all_tables("test_db"):
        hive_client.drop_table("test_db", "test_table", True)
    hive_client.create_table(table)

    for i in range(1, 6):
        partition = ttypes.Partition(
            dbName="test_db",
            tableName="test_table",
            values=[str(i)],
            sd=storageDesc,
            parameters={},
        )
        hive_client.add_partition(partition)

    # testing primitive Types
    cols2 = [
        ttypes.FieldSchema(name="col3", type="void", comment="c3"),
        ttypes.FieldSchema(name="col4", type="boolean", comment="c4"),
        ttypes.FieldSchema(name="col5", type="tinyint", comment="c5"),
        ttypes.FieldSchema(name="col6", type="smallint", comment="c6"),
        ttypes.FieldSchema(name="col7", type="bigint", comment="c7"),
        ttypes.FieldSchema(name="col8", type="float", comment="c8"),
        ttypes.FieldSchema(name="col9", type="double", comment="c9"),
        ttypes.FieldSchema(name="col10", type="date", comment="c10"),
        ttypes.FieldSchema(name="col11", type="timestamp", comment="c11"),
        ttypes.FieldSchema(
            name="col12", type="timestamp with local time zone", comment="c12"
        ),
        ttypes.FieldSchema(name="col13", type="interval_year_month", comment="c13"),
        ttypes.FieldSchema(name="col14", type="interval_day_time", comment="c14"),
        ttypes.FieldSchema(name="col15", type="binary", comment="c15"),
    ]

    storageDesc = ttypes.StorageDescriptor(
        location="/tmp/test_db/test_table2",
        cols=cols2,
        inputFormat="org.apache.hadoop.mapred.TextInputFormat",
        outputFormat="org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat",
        compressed=False,
        numBuckets=-1,
        serdeInfo=serde_info,
        bucketCols=[],
    )

    table = ttypes.Table(
        tableName="test_table2",
        dbName="test_db",
        owner="owner",
        createTime=0,
        lastAccessTime=0,
        retention=0,
        sd=storageDesc,
        partitionKeys=partitionKeys,
        parameters={},
        tableType="EXTERNAL_TABLE",
    )

    if "test_table2" in hive_client.get_all_tables("test_db"):
        hive_client.drop_table("test_db", "test_table2", True)
    hive_client.create_table(table)

    # testing Parameterized Types
    cols3 = [
        ttypes.FieldSchema(name="col16", type="decimal(10,2)", comment="c16"),
        ttypes.FieldSchema(name="col17", type="varchar(10)", comment="c17"),
        ttypes.FieldSchema(name="col18", type="char(10)", comment="c18"),
        ttypes.FieldSchema(name="col19", type="array<int>", comment="c19"),
        ttypes.FieldSchema(name="col20", type="map<int,string>", comment="c20"),
        ttypes.FieldSchema(name="col21", type="struct<a:int,b:string>", comment="c21"),
        ttypes.FieldSchema(name="col22", type="uniontype<int,string>", comment="c22"),
    ]

    storageDesc = ttypes.StorageDescriptor(
        location="/tmp/test_db/test_table3",
        cols=cols3,
        inputFormat="org.apache.hadoop.mapred.TextInputFormat",
        outputFormat="org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat",
        compressed=False,
        numBuckets=-1,
        serdeInfo=serde_info,
        bucketCols=[],
    )

    table = ttypes.Table(
        tableName="test_table3",
        dbName="test_db",
        owner="owner",
        createTime=0,
        lastAccessTime=0,
        retention=0,
        sd=storageDesc,
        partitionKeys=partitionKeys,
        parameters={},
        tableType="EXTERNAL_TABLE",
    )

    if "test_table3" in hive_client.get_all_tables("test_db"):
        hive_client.drop_table("test_db", "test_table3", True)
    hive_client.create_table(table)


def test_list_databases(hive_client):
    assert "test_db" in HMS(hive_client).list_databases()


def test_get_database(hive_client):
    hms = HMS(hive_client)
    database = hms.get_database("test_db")
    assert isinstance(database, HDatabase)
    assert database.name == "test_db"
    assert database.location == "file:/tmp/test_db.db"
    assert database.owner_name == "owner"


def test_list_tables(hive_client):
    hms = HMS(hive_client)
    tables = hms.list_tables("test_db")
    assert isinstance(tables, list)
    assert "test_table" in tables


@pytest.mark.xfail(reason="'table_type' is coming back MANAGED_TABLE.")
def test_get_table(hive_client):
    hms = HMS(hive_client)
    table = hms.get_table("test_db", "test_table")
    assert isinstance(table, HTable)
    assert table.database_name == "test_db"
    assert table.name == "test_table"
    assert table.owner == "owner"
    assert table.storage.location == "file:/tmp/test_db/test_table"
    assert len(table.columns) == 2
    assert isinstance(table.columns[0], HColumn)
    assert isinstance(table.columns[1], HColumn)
    assert len(table.partition_columns) == 1
    assert isinstance(table.partition_columns[0], HColumn)
    assert isinstance(table.parameters, dict)
    assert len(table.parameters) == 1
    assert (
        table.parameters.get("transient_lastDdlTime") is not None
    )  # this is not a parameter of the table we created, but the metastore adds it
    # This assertion fails, I leave it here on purpose. My current assumption is that
    # the metastore overrides some of the passed options with defaults. "MANAGEd_TABLE"
    # is the default value for tableType. See here for the defaults:
    # https://github.com/apache/hive/blob/14a1f70607db5ae6cf71b6d4343f308a5167581c/standalone-metastore/metastore-server/src/main/java/org/apache/hadoop/hive/metastore/client/builder/TableBuilder.java#L69C43-L69C43
    # Something similar happens also when you choose a "VIRTUAL_VIEW" table type, where
    # the metastore overrides the storage descriptor removing in the location the file::
    # prefix.
    assert table.table_type == "EXTERNAL_TABLE"


def test_get_table_columns(hive_client):
    hms = HMS(hive_client)
    table = hms.get_table("test_db", "test_table")
    columns = table.columns
    partition_columns = table.partition_columns

    assert columns[0].name == "col1"
    assert isinstance(columns[0].type, HType)
    assert isinstance(columns[0].type, HPrimitiveType)
    assert columns[0].type.name == "INT"
    assert columns[0].type.category == HTypeCategory.PRIMITIVE
    assert columns[0].comment == "c1"

    assert columns[1].name == "col2"
    assert isinstance(columns[1].type, HType)
    assert isinstance(columns[1].type, HPrimitiveType)
    assert columns[1].type.name == "STRING"
    assert columns[1].type.category == HTypeCategory.PRIMITIVE
    assert columns[1].comment == "c2"

    assert partition_columns[0].name == "partition"
    assert isinstance(partition_columns[0].type, HType)
    assert isinstance(partition_columns[0].type, HPrimitiveType)
    assert partition_columns[0].type.name == "STRING"
    assert partition_columns[0].type.category == HTypeCategory.PRIMITIVE
    assert partition_columns[0].comment == ""


def test_list_columns(hive_client):
    hms = HMS(hive_client)
    columns = hms.list_columns("test_db", "test_table")

    assert isinstance(columns, list)
    assert len(columns) == 2
    assert columns[0] == "col1"
    assert columns[1] == "col2"


def test_list_partitions(hive_client):
    hms = HMS(hive_client)
    partitions = hms.list_partitions("test_db", "test_table")

    assert isinstance(partitions, list)
    assert len(partitions) == 5
    for i in range(1, 6):
        assert f"partition={i}" in partitions


def test_get_partitions(hive_client):
    hms = HMS(hive_client)
    partitions = hms.get_partitions("test_db", "test_table")
    expected_storage = HStorage(
        storage_format=StorageFormat(
            serde="org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe",
            input_format="org.apache.hadoop.mapred.TextInputFormat",
            output_format="org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat",
        ),
        skewed=True,
        location="file:/tmp/test_db/test_table",
        bucket_property=HiveBucketProperty(
            bucketed_by=[],
            bucket_count=-1,
            version=BucketingVersion.V1,
            sorting_columns=[],
        ),
        serde_parameters={"field.delim": ","},
    )

    assert isinstance(partitions, list)
    assert len(partitions) == 5  # we created 5 partitions in the setup_data fixture

    missing_partitions = {1, 2, 3, 4, 5}

    for partition in partitions:
        assert isinstance(partition, HPartition)
        missing_partitions.discard(int(partition.values[0]))
        assert partition.database_name == "test_db"
        assert partition.table_name == "test_table"
        assert partition.sd == expected_storage
        assert partition.create_time != 0
        assert partition.last_access_time == 0
        assert partition.cat_name == "hive"
        assert isinstance(partition.parameters, dict)
        assert partition.write_id == -1

    assert missing_partitions == set()


def test_get_partition(hive_client):
    hms = HMS(hive_client)
    partition = hms.get_partition("test_db", "test_table", "partition=1")
    expected_storage = HStorage(
        storage_format=StorageFormat(
            serde="org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe",
            input_format="org.apache.hadoop.mapred.TextInputFormat",
            output_format="org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat",
        ),
        skewed=True,
        location="file:/tmp/test_db/test_table",
        bucket_property=HiveBucketProperty(
            bucketed_by=[],
            bucket_count=-1,
            version=BucketingVersion.V1,
            sorting_columns=[],
        ),
        serde_parameters={"field.delim": ","},
    )

    assert isinstance(partition, HPartition)
    assert partition.database_name == "test_db"
    assert partition.table_name == "test_table"
    assert partition.values == ["1"]
    assert partition.sd == expected_storage
    assert partition.create_time != 0
    assert partition.last_access_time == 0
    assert partition.cat_name == "hive"
    assert isinstance(partition.parameters, dict)
    assert partition.write_id == -1


def test_primitive_types(hive_client):
    hms = HMS(hive_client)

    table = hms.get_table("test_db", "test_table2")
    columns = table.columns

    assert len(columns) == 13

    assert columns[0].name == "col3"
    assert isinstance(columns[0].type, HType)
    assert isinstance(columns[0].type, HPrimitiveType)
    assert columns[0].type.name == "VOID"
    assert columns[0].type.category == HTypeCategory.PRIMITIVE
    assert columns[0].comment == "c3"

    assert columns[1].name == "col4"
    assert isinstance(columns[1].type, HType)
    assert isinstance(columns[1].type, HPrimitiveType)
    assert columns[1].type.name == "BOOLEAN"
    assert columns[1].type.category == HTypeCategory.PRIMITIVE
    assert columns[1].comment == "c4"

    assert columns[2].name == "col5"
    assert isinstance(columns[2].type, HType)
    assert isinstance(columns[2].type, HPrimitiveType)
    assert columns[2].type.name == "BYTE"
    assert columns[2].type.category == HTypeCategory.PRIMITIVE
    assert columns[2].comment == "c5"

    assert columns[3].name == "col6"
    assert isinstance(columns[3].type, HType)
    assert isinstance(columns[3].type, HPrimitiveType)
    assert columns[3].type.name == "SHORT"
    assert columns[3].type.category == HTypeCategory.PRIMITIVE
    assert columns[3].comment == "c6"

    assert columns[4].name == "col7"
    assert isinstance(columns[4].type, HType)
    assert isinstance(columns[4].type, HPrimitiveType)
    assert columns[4].type.name == "LONG"
    assert columns[4].type.category == HTypeCategory.PRIMITIVE
    assert columns[4].comment == "c7"

    assert columns[5].name == "col8"
    assert isinstance(columns[5].type, HType)
    assert isinstance(columns[5].type, HPrimitiveType)
    assert columns[5].type.name == "FLOAT"
    assert columns[5].type.category == HTypeCategory.PRIMITIVE
    assert columns[5].comment == "c8"

    assert columns[6].name == "col9"
    assert isinstance(columns[6].type, HType)
    assert isinstance(columns[6].type, HPrimitiveType)
    assert columns[6].type.name == "DOUBLE"
    assert columns[6].type.category == HTypeCategory.PRIMITIVE
    assert columns[6].comment == "c9"

    assert columns[7].name == "col10"
    assert isinstance(columns[7].type, HType)
    assert isinstance(columns[7].type, HPrimitiveType)
    assert columns[7].type.name == "DATE"
    assert columns[7].type.category == HTypeCategory.PRIMITIVE
    assert columns[7].comment == "c10"

    assert columns[8].name == "col11"
    assert isinstance(columns[8].type, HType)
    assert isinstance(columns[8].type, HPrimitiveType)
    assert columns[8].type.name == "TIMESTAMP"
    assert columns[8].type.category == HTypeCategory.PRIMITIVE
    assert columns[8].comment == "c11"

    assert columns[9].name == "col12"
    assert isinstance(columns[9].type, HType)
    assert isinstance(columns[9].type, HPrimitiveType)
    assert columns[9].type.name == "TIMESTAMPLOCALTZ"
    assert columns[9].type.category == HTypeCategory.PRIMITIVE
    assert columns[9].comment == "c12"

    assert columns[10].name == "col13"
    assert isinstance(columns[10].type, HType)
    assert isinstance(columns[10].type, HPrimitiveType)
    assert columns[10].type.name == "INTERVAL_YEAR_MONTH"
    assert columns[10].type.category == HTypeCategory.PRIMITIVE
    assert columns[10].comment == "c13"

    assert columns[11].name == "col14"
    assert isinstance(columns[11].type, HType)
    assert isinstance(columns[11].type, HPrimitiveType)
    assert columns[11].type.name == "INTERVAL_DAY_TIME"
    assert columns[11].type.category == HTypeCategory.PRIMITIVE
    assert columns[11].comment == "c14"

    assert columns[12].name == "col15"
    assert isinstance(columns[12].type, HType)
    assert isinstance(columns[12].type, HPrimitiveType)
    assert columns[12].type.name == "BINARY"
    assert columns[12].type.category == HTypeCategory.PRIMITIVE
    assert columns[12].comment == "c15"


def test_parameterized_types(hive_client):
    hms = HMS(hive_client)
    table = hms.get_table("test_db", "test_table3")
    columns = table.columns

    assert len(columns) == 7

    assert columns[0].name == "col16"
    assert isinstance(columns[0].type, HType)
    assert isinstance(columns[0].type, HDecimalType)
    assert columns[0].type.name == "DECIMAL"
    assert columns[0].type.category == HTypeCategory.PRIMITIVE
    assert columns[0].comment == "c16"
    assert columns[0].type.precision == 10
    assert columns[0].type.scale == 2

    assert columns[1].name == "col17"
    assert isinstance(columns[1].type, HType)
    assert isinstance(columns[1].type, HVarcharType)
    assert columns[1].type.name == "VARCHAR"
    assert columns[1].type.category == HTypeCategory.PRIMITIVE
    assert columns[1].comment == "c17"
    assert columns[1].type.length == 10

    assert columns[2].name == "col18"
    assert isinstance(columns[2].type, HType)
    assert isinstance(columns[2].type, HCharType)
    assert columns[2].type.name == "CHAR"
    assert columns[2].type.category == HTypeCategory.PRIMITIVE
    assert columns[2].comment == "c18"
    assert columns[2].type.length == 10

    assert columns[3].name == "col19"
    assert isinstance(columns[3].type, HType)
    assert isinstance(columns[3].type, HListType)
    assert columns[3].type.name == "LIST"
    assert columns[3].type.category == HTypeCategory.LIST
    assert columns[3].comment == "c19"
    assert isinstance(columns[3].type.element_type, HType)
    assert isinstance(columns[3].type.element_type, HPrimitiveType)
    assert columns[3].type.element_type.name == "INT"

    assert columns[4].name == "col20"
    assert isinstance(columns[4].type, HType)
    assert isinstance(columns[4].type, HMapType)
    assert columns[4].type.name == "MAP"
    assert columns[4].type.category == HTypeCategory.MAP
    assert columns[4].comment == "c20"
    assert isinstance(columns[4].type.key_type, HType)
    assert isinstance(columns[4].type.key_type, HPrimitiveType)
    assert columns[4].type.key_type.name == "INT"
    assert isinstance(columns[4].type.value_type, HType)
    assert isinstance(columns[4].type.value_type, HPrimitiveType)
    assert columns[4].type.value_type.name == "STRING"

    assert columns[5].name == "col21"
    assert isinstance(columns[5].type, HType)
    assert isinstance(columns[5].type, HStructType)
    assert columns[5].type.name == "STRUCT"
    assert columns[5].type.category == HTypeCategory.STRUCT
    assert columns[5].comment == "c21"
    assert len(columns[5].type.names) == 2
    assert len(columns[5].type.types) == 2
    assert columns[5].type.names[0] == "a"
    assert columns[5].type.names[1] == "b"
    assert isinstance(columns[5].type.types[0], HType)
    assert isinstance(columns[5].type.types[0], HPrimitiveType)
    assert columns[5].type.types[0].name == "INT"
    assert isinstance(columns[5].type.types[1], HType)
    assert isinstance(columns[5].type.types[1], HPrimitiveType)
    assert columns[5].type.types[1].name == "STRING"

    assert columns[6].name == "col22"
    assert isinstance(columns[6].type, HType)
    assert isinstance(columns[6].type, HUnionType)
    assert columns[6].type.name == "UNION"
    assert columns[6].type.category == HTypeCategory.UNION
    assert columns[6].comment == "c22"
    assert len(columns[6].type.types) == 2
    assert isinstance(columns[6].type.types[0], HType)
    assert isinstance(columns[6].type.types[0], HPrimitiveType)
    assert columns[6].type.types[0].name == "INT"
    assert isinstance(columns[6].type.types[1], HType)
    assert isinstance(columns[6].type.types[1], HPrimitiveType)
    assert columns[6].type.types[1].name == "STRING"



from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional

from thrift.protocol.TBinaryProtocol import TBinaryProtocol
from thrift.transport import TSocket, TTransport

from .hive_metastore import ThriftHiveMetastore as hms
from .hive_metastore.ttypes import (
    Database,
    FieldSchema,
    Partition,
    PrincipalType,
    SerDeInfo,
    StorageDescriptor,
    Table,
)
from .htypes import HType, TypeParser


class HPrincipalType(Enum):
    ROLE = "ROLE"
    USER = "USER"
    GROUP = "GROUP"


@dataclass
class HPrivilegeGrantInfo:
    privilege: str
    grantor: str
    grantor_type: HPrincipalType
    create_time: int
    grant_option: bool


@dataclass
class HPrincipalPrivilegeSet:
    user_privileges: List[HPrivilegeGrantInfo]
    group_privileges: List[HPrivilegeGrantInfo]
    role_privileges: List[HPrivilegeGrantInfo]


@dataclass
class HDatabase:
    name: str
    location: Optional[str] = None
    owner_name: Optional[str] = None
    owner_type: Optional[HPrincipalType] = None
    comment: Optional[str] = None
    parameters: Optional[Dict[str, str]] = None


@dataclass
class HColumn:
    name: str
    type: HType
    comment: Optional[str] = None


class HSortingOrder(Enum):
    ASC = 1
    DESC = 0


@dataclass
class HSortingColumn:
    column: str
    order: HSortingOrder


class BucketingVersion(Enum):
    V1 = 1
    V2 = 2


@dataclass
class HiveBucketProperty:
    bucketed_by: List[str]
    bucket_count: int
    version: BucketingVersion = BucketingVersion.V1
    sorting_columns: List[HSortingColumn] = field(default_factory=list)


@dataclass
class StorageFormat:
    serde: str
    input_format: str
    output_format: str


@dataclass
class HStorage:
    storage_format: StorageFormat
    skewed: bool = False
    location: Optional[str] = None
    bucket_property: Optional[HiveBucketProperty] = None
    serde_parameters: Optional[Dict[str, str]] = None


@dataclass
class HTable:
    database_name: str
    name: str
    table_type: str
    columns: List[HColumn]
    partition_columns: List[HColumn]
    storage: HStorage
    parameters: Dict[str, str]
    view_original_text: Optional[str] = None
    view_expanded_text: Optional[str] = None
    write_id: Optional[int] = None
    owner: Optional[str] = None


@dataclass
class HSkewedInfo:
    skewed_col_names: List[str]
    skewed_col_values: List[List[str]]
    skewed_col_value_location_maps: Dict[List[str], str]


@dataclass
class HPartition:
    database_name: str
    table_name: str
    values: List[str]
    parameters: Dict[str, str]
    create_time: int
    last_access_time: int
    sd: HStorage
    cat_name: str
    write_id: int


class HMS:
    def __init__(self, client: hms.Client):
        self.client = client

    @staticmethod
    def create(host: str = "localhost", port: int = 9083) -> "_HMSConnection":
        return _HMSConnection(host, port)

    def list_databases(self) -> List[str]:
        databases = self.client.get_all_databases()
        db_names = []
        for database in databases:
            db_names.append(database)
        return db_names

    def get_database(self, name: str) -> HDatabase:
        db: Database = self.client.get_database(name)

        if db.ownerType is PrincipalType.USER:
            owner_type = HPrincipalType.USER
        elif db.ownerType is PrincipalType.ROLE:
            owner_type = HPrincipalType.ROLE
        else:
            owner_type = None

        return HDatabase(
            db.name,  # pyright: ignore[reportGeneralTypeIssues]
            db.locationUri,
            db.ownerName,
            owner_type,
            db.description,
            db.parameters,
        )

    def list_tables(self, database_name: str) -> List[str]:
        return self.client.get_all_tables(database_name)

    def list_columns(self, database_name: str, table_name: str) -> List[str]:
        # TODO: Rather than ignore these pyright errors, do appropriate None handling
        columns = self.client.get_table(
            database_name,
            table_name,
        ).sd.cols  # pyright: ignore[reportOptionalMemberAccess]
        self.client.get_schema(database_name, table_name)
        column_names = []
        for column in columns:  # pyright: ignore[reportOptionalIterable]
            column_names.append(column.name)
        return column_names

    def list_partitions(
        self,
        database_name: str,
        table_name: str,
        max_parts: int = -1,
    ) -> List[str]:
        partitions = self.client.get_partition_names(
            database_name,
            table_name,
            max_parts,
        )
        return partitions

    def get_partitions(
        self,
        database_name: str,
        table_name: str,
        max_parts: int = -1,
    ) -> List[HPartition]:
        result_partitions = []
        partitions: List[Partition] = self.client.get_partitions(
            database_name,
            table_name,
            max_parts,
        )

        assert isinstance(partitions, list)

        for partition in partitions:
            assert isinstance(partition, Partition)
            assert isinstance(partition.sd, StorageDescriptor)
            assert isinstance(partition.sd.serdeInfo, SerDeInfo)
            assert isinstance(partition.dbName, str)
            assert isinstance(partition.tableName, str)
            assert isinstance(partition.values, list)
            assert isinstance(partition.parameters, dict)
            assert isinstance(partition.createTime, int)
            assert isinstance(partition.lastAccessTime, int)
            assert isinstance(partition.catName, str)

            serialization_lib = (
                ""
                if partition.sd.serdeInfo.serializationLib is None
                else partition.sd.serdeInfo.serializationLib
            )
            input_format = (
                "" if partition.sd.inputFormat is None else partition.sd.inputFormat
            )
            output_format = (
                "" if partition.sd.outputFormat is None else partition.sd.outputFormat
            )
            bucket_cols = (
                [] if partition.sd.bucketCols is None else partition.sd.bucketCols
            )
            sort_cols = [] if partition.sd.sortCols is None else partition.sd.sortCols
            num_buckets = (
                0 if partition.sd.numBuckets is None else partition.sd.numBuckets
            )
            is_skewed = partition.sd.skewedInfo is not None

            assert isinstance(serialization_lib, str)
            assert isinstance(input_format, str)
            assert isinstance(output_format, str)
            assert isinstance(bucket_cols, list)
            assert isinstance(sort_cols, list)
            assert isinstance(num_buckets, int)

            storage_format = StorageFormat(
                serialization_lib,
                input_format,
                output_format,
            )

            bucket_property = HiveBucketProperty(
                bucket_cols,
                num_buckets,
                BucketingVersion.V1,
                sort_cols,
            )

            sd = HStorage(
                storage_format,
                is_skewed,
                partition.sd.location,
                bucket_property,
                partition.sd.serdeInfo.parameters,
            )

            result_partition = HPartition(
                partition.dbName,
                partition.tableName,
                partition.values,
                partition.parameters,
                partition.createTime,
                partition.lastAccessTime,
                sd,
                partition.catName,
                partition.writeId,
            )

            result_partitions.append(result_partition)

        return result_partitions

    def get_partition(
        self,
        database_name: str,
        table_name: str,
        partition_name: str,
    ) -> HPartition:
        partition: Partition = self.client.get_partition_by_name(
            database_name,
            table_name,
            partition_name,
        )
        assert partition is not None
        assert isinstance(partition, Partition)
        assert isinstance(partition.sd, StorageDescriptor)
        assert isinstance(partition.sd.serdeInfo, SerDeInfo)
        assert isinstance(partition.sd.serdeInfo.serializationLib, str)
        assert partition.sd.inputFormat is not None
        assert partition.sd.outputFormat is not None

        serialization_lib = partition.sd.serdeInfo.serializationLib
        input_format = partition.sd.inputFormat
        output_format = partition.sd.outputFormat
        storage_format = StorageFormat(
            serialization_lib,
            input_format,
            output_format,
        )
        sort_cols = [] if partition.sd.sortCols is None else partition.sd.sortCols
        bucket_cols = [] if partition.sd.bucketCols is None else partition.sd.bucketCols
        num_buckets = partition.sd.numBuckets or 0
        is_skewed = partition.sd.skewedInfo is not None
        location = "" if partition.sd.location is None else partition.sd.location
        serde_parameters = (
            {}
            if partition.sd.serdeInfo.parameters is None
            else partition.sd.serdeInfo.parameters
        )
        catName = "" if partition.catName is None else partition.catName
        writeId = -1 if partition.writeId is None else partition.writeId
        last_access_time = (
            -1 if partition.lastAccessTime is None else partition.lastAccessTime
        )
        partition_parameters = (
            {} if partition.parameters is None else partition.parameters
        )
        create_time = -1 if partition.createTime is None else partition.createTime
        values = [] if partition.values is None else partition.values
        db_name = partition.dbName
        partition_table_name = partition.tableName

        assert isinstance(sort_cols, list)
        assert isinstance(bucket_cols, list)
        assert isinstance(num_buckets, int)
        assert isinstance(location, str)
        assert isinstance(serde_parameters, dict)
        assert isinstance(catName, str)
        assert isinstance(writeId, int)
        assert isinstance(last_access_time, int)
        assert isinstance(partition_parameters, dict)
        assert isinstance(create_time, int)
        assert isinstance(values, list)
        assert isinstance(db_name, str)
        assert isinstance(partition_table_name, str)

        bucket_property = HiveBucketProperty(
            bucket_cols,
            num_buckets,
            BucketingVersion.V1,
            sort_cols,
        )
        sd = HStorage(
            storage_format,
            is_skewed,
            location,
            bucket_property,
            serde_parameters,
        )
        result_partition = HPartition(
            db_name,
            partition_table_name,
            values,
            partition_parameters,
            create_time,
            last_access_time,
            sd,
            catName,
            writeId,
        )

        return result_partition

    def get_table(self, database_name: str, table_name: str) -> HTable:
        table: Table = self.client.get_table(database_name, table_name)

        columns = []

        partition_columns = []
        if table.partitionKeys is not None:
            if isinstance(table.partitionKeys, list):
                t_part_columns: List[FieldSchema] = table.partitionKeys
                for column in t_part_columns:
                    if column is not None:
                        if isinstance(column, FieldSchema):
                            if column.type is not None:
                                type_parser = TypeParser(column.type)
                            else:
                                raise TypeError(f"Expected type to be str, got None")
                            if column.comment is not None:
                                comment = column.comment
                            else:
                                comment = ""
                            if column.name is not None:
                                name = column.name
                            else:
                                raise TypeError(f"Expected name to be str, got None")
                            partition_columns.append(
                                HColumn(name, type_parser.parse_type(), comment)
                            )

        if table.sd is not None:
            if table.sd.cols is not None:
                if isinstance(table.sd.cols, list):
                    t_columns: List[FieldSchema] = table.sd.cols
                    for column in t_columns:
                        if column is not None:
                            if isinstance(column, FieldSchema):
                                if column.type is not None:
                                    type_parser = TypeParser(column.type)
                                else:
                                    raise TypeError(
                                        f"Expected type to be str, got None"
                                    )
                                if column.comment is not None:
                                    comment = column.comment
                                else:
                                    comment = ""
                                if column.name is not None:
                                    name = column.name
                                else:
                                    raise TypeError(
                                        f"Expected name to be str, got None"
                                    )
                                columns.append(
                                    HColumn(name, type_parser.parse_type(), comment)
                                )

            if table.sd.serdeInfo is not None:
                if isinstance(table.sd.serdeInfo, SerDeInfo):
                    if table.sd.serdeInfo.serializationLib is not None:
                        if isinstance(table.sd.serdeInfo.serializationLib, str):
                            serde = table.sd.serdeInfo.serializationLib
                        else:
                            raise TypeError(
                                f"Expected serializationLib to be str, got {type(table.sd.serdeInfo.serializationLib)}"
                            )
                    else:
                        raise TypeError(
                            f"Expected serdeInfo to be str, got {type(table.sd.serdeInfo)}"
                        )
                else:
                    raise TypeError(
                        f"Expected serdeInfo to be SerDeInfo, got {type(table.sd.serdeInfo)}"
                    )
            else:
                raise TypeError(f"Expected serdeInfo to be SerDeInfo, got None")

            if table.sd.inputFormat is not None:
                if isinstance(table.sd.inputFormat, str):
                    input_format = table.sd.inputFormat
                else:
                    raise TypeError(
                        f"Expected inputFormat to be str, got {type(table.sd.inputFormat)}"
                    )
            else:
                raise TypeError(f"Expected inputFormat to be str, got None")

            if table.sd.outputFormat is not None:
                if isinstance(table.sd.outputFormat, str):
                    output_format = table.sd.outputFormat
                else:
                    raise TypeError(
                        f"Expected outputFormat to be str, got {type(table.sd.outputFormat)}"
                    )
            else:
                raise TypeError(f"Expected outputFormat to be str, got None")

            storage_format = StorageFormat(serde, input_format, output_format)

            bucket_property = None
            if table.sd.bucketCols is not None:
                sort_cols = []
                if table.sd.sortCols is not None:
                    if isinstance(table.sd.sortCols, list):
                        for order in table.sd.sortCols:
                            sort_cols.append(HSortingColumn(order.col, order.order))
                    else:
                        raise TypeError(
                            f"Expected bucketCols to be list, got {type(table.sd.sortCols)}"
                        )

                version = BucketingVersion.V1
                if table.parameters is not None:
                    if isinstance(table.parameters, dict):
                        if (
                            table.parameters.get(
                                "TABLE_BUCKETING_VERSION", BucketingVersion.V1
                            )
                            == BucketingVersion.V2
                        ):
                            version = BucketingVersion.V2
                    else:
                        raise TypeError(
                            f"Expected parameters to be dict, got {type(table.parameters)}"
                        )
                else:
                    raise TypeError(
                        f"Expected parameters to be dict, got {type(table.parameters)}"
                    )

                if table.sd.numBuckets is not None:
                    if isinstance(table.sd.numBuckets, int):
                        num_buckets = table.sd.numBuckets
                    else:
                        raise TypeError(
                            f"Expected numBuckets to be int, got {type(table.sd.numBuckets)}"
                        )
                else:
                    raise TypeError(
                        f"Expected numBuckets to be int, got {type(table.sd.numBuckets)}"
                    )

                bucket_property = HiveBucketProperty(
                    table.sd.bucketCols, num_buckets, version, sort_cols
                )

            if table.sd.skewedInfo is None:
                is_skewed = False
            else:
                is_skewed = True

            if table.sd.location is not None:
                if isinstance(table.sd.location, str):
                    location = table.sd.location
                else:
                    raise TypeError(
                        f"Expected location to be str, got {type(table.sd.location)}"
                    )
            else:
                location = None

            if table.sd.serdeInfo is not None:
                if isinstance(table.sd.serdeInfo, SerDeInfo):
                    serde_info = table.sd.serdeInfo
                else:
                    raise TypeError(
                        f"Expected serdeInfo to be SerDeInfo, got {type(table.sd.serdeInfo)}"
                    )
            else:
                raise TypeError(
                    f"Expected serdeInfo to be SerDeInfo, got {type(table.sd.serdeInfo)}"
                )
            if serde_info.parameters is not None:
                if isinstance(serde_info.parameters, dict):
                    serde_parameters = serde_info.parameters
                else:
                    raise TypeError(
                        f"Expected serdeInfo.parameters to be dict, got {type(serde_info.parameters)}"
                    )
            else:
                raise TypeError(
                    f"Expected serdeInfo.parameters to be dict, got {type(serde_info.parameters)}"
                )
        else:
            raise TypeError(
                f"Expected sd to be StorageDescriptor, got {type(table.sd)}"
            )

        storage = HStorage(
            storage_format,
            is_skewed,
            location,
            bucket_property,
            serde_parameters,
        )

        if table.parameters is not None:
            if isinstance(table.parameters, dict):
                params = table.parameters
            else:
                raise TypeError(
                    f"Expected parameters to be dict, got {type(table.parameters)}"
                )
        else:
            raise TypeError(
                f"Expected parameters to be dict, got {type(table.parameters)}"
            )

        if table.tableType is not None:
            if isinstance(table.tableType, str):
                table_type = table.tableType
            else:
                raise TypeError(
                    f"Expected tableType to be str, got {type(table.tableType)}"
                )
        else:
            raise TypeError(
                f"Expected tableType to be str, got {type(table.tableType)}"
            )

        if table.tableName is not None:
            table_name = table.tableName
        else:
            raise TypeError(
                f"Expected tableName to be str, got {type(table.tableName)}"
            )

        if table.dbName is not None:
            db_name = table.dbName
        else:
            raise TypeError(f"Expected dbName to be str, got {type(table.dbName)}")

        return HTable(
            db_name,
            table_name,
            table_type,
            columns,
            partition_columns,
            storage,
            params,
            table.viewOriginalText,
            table.viewExpandedText,
            table.writeId,
            table.owner,
        )


class _HMSConnection:
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.transport = None

    def __enter__(self):
        socket = TSocket.TSocket(self.host, self.port)
        self.transport = TTransport.TBufferedTransport(socket)
        protocol = TBinaryProtocol(self.transport)
        self.transport.open()
        return HMS(hms.Client(protocol))

    def __exit__(self, type, value, traceback):
        if self.transport is not None:
            self.transport.close()
